#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <libndls.h>

// you should set this to the expected amount of objects that get allocated, so the wrapper doesn't need to resize the array as often
#define BUCKETS_START_SIZE 20
#define LIST_START_SIZE 20
#define BUCKET_GRANULARITY 8
// the region of a pointer is p >> BUCKET_GRANULARITY
// BUCKET_GRANULARITY should be smaller if you use smaller memory blocks


//#define DEBUG 1




extern "C"
{
	// we want this to be as fast as possible
	#pragma GCC optimize "-O3"
	#ifdef DEBUG
		static uint32_t malloc_count = 0; // should be 0 after all is freed
	#endif
	struct bucket {
		uint32_t region; // for which region of memory this bucket is responsible. 0 indicates an unused bucket
		void** list; // list of allocations
		uint32_t size; // size of the list
	};
	//static void** memory_allocations = NULL;
	//static uint32_t memory_allocations_size = 0;
	static struct bucket* buckets = NULL;
	static uint32_t buckets_size = 0;
	void* __real_malloc(size_t size);
	void __real_free(void* p);
	
	static bool bucket_add_allocation(struct bucket* bucket,void* p)
	{
		if (bucket->list == NULL)
		{
			bucket->list = (void**) __real_malloc(sizeof(void*)*LIST_START_SIZE);
			if (bucket->list == NULL)
			{
				return false;
			}
			bucket->size = LIST_START_SIZE;
		}
		for (uint32_t i = 0;i<bucket->size;i++)
		{
			if ((bucket->list)[i] == NULL)
			{
				(bucket->list)[i] = p;
				return true;
			}
		}
		// no free entry
		void** new_bucket_list = (void**) malloc(sizeof(void*)*bucket->size*2);
		if (new_bucket_list == NULL)
		{
			return false;
		}
		for (uint32_t i = 0;i<bucket->size*2;i++)
		{
			new_bucket_list[i] = NULL;
		}
		for (uint32_t i = 0;i<bucket->size;i++)
		{
			new_bucket_list[i] = (bucket->list)[i];
		}
		__real_free(bucket->list);
		bucket->list = new_bucket_list;
		bucket->size *= 2;
		for (uint32_t i = 0;i<bucket->size;i++)
		{
			if ((bucket->list)[i] == NULL)
			{
				(bucket->list)[i] = p;
				return true;
			}
		}
		return false;
	}
	static void bucket_remove_allocation(struct bucket* bucket,void* p)
	{
		for (uint32_t i = 0;i<bucket->size;i++) // also handles list==NULL, because then size is 0
		{
			if ((bucket->list)[i] == p)
			{
				(bucket->list)[i] = NULL;
				return;
			}
		}
	}
	static bool add_allocation(void* p)
	{
		uint32_t region = (uint32_t) p >> BUCKET_GRANULARITY;
		if (buckets == NULL)
		{
			buckets = (struct bucket*) __real_malloc(sizeof(struct bucket)*BUCKETS_START_SIZE);
			if (buckets != NULL)
			{
				buckets_size = BUCKETS_START_SIZE;
				for (uint32_t i = 0;i<buckets_size;i++)
				{
					buckets[i] = {0,NULL,0};
				}
			}
			else
			{
				return false;
			}
		}
		for (uint32_t i = 0;i<buckets_size;i++)
		{
			if (buckets[i].region == region)
			{
				if (bucket_add_allocation(buckets+i,p))
				{
					return true;
				}
				else
				{
					return false;
				}
			}
		}
		// no bucket with the region exists
		for (uint32_t i = 0;i<buckets_size;i++)
		{
			if (buckets[i].region == 0)
			{
				buckets[i].region = region;
				if (bucket_add_allocation(buckets+i,p))
				{
					return true;
				}
				else
				{
					return false;
				}
			}
		}
		// no empty buckets
		struct bucket* new_buckets = (struct bucket*) __real_malloc(sizeof(struct bucket)*buckets_size*2);
		if (new_buckets == NULL)
		{
			return false;
		}
		for (uint32_t i = 0;i<buckets_size*2;i++)
		{
			new_buckets[i] = {0,NULL,0};
		}
		for (uint32_t i = 0;i<buckets_size;i++)
		{
			new_buckets[i] = buckets[i];
		}
		__real_free(buckets);
		buckets = new_buckets;
		buckets_size *= 2;
		for (uint32_t i = 0;i<buckets_size;i++)
		{
			if (buckets[i].region == 0)
			{
				buckets[i].region = region;
				if (bucket_add_allocation(buckets+i,p))
				{
					return true;
				}
				else
				{
					return false;
				}
			}
		}
		return false;
	}
	static void remove_allocation(void* p)
	{
		uint32_t region = (uint32_t) p >> BUCKET_GRANULARITY;
		for (uint32_t i = 0;i<buckets_size;i++)
		{
			if (buckets[i].region == region)
			{
				bucket_remove_allocation(buckets+i,p);
				return;
			}
		}
	}
	
	void* __attribute__((malloc)) __wrap_malloc(size_t size)
	{
		void* block = __real_malloc(size);
		if (block == NULL)
		{
			return NULL;
		}
		if (add_allocation(block))
		{
			return block;
		}
		__real_free(block);
		return NULL;
	}
	void __wrap_free(void* p)
	{
		#ifdef DEBUG
			puts("free called!");
			fflush(stdout);
		#endif
		remove_allocation(p);
		#ifdef DEBUG
			malloc_count--;
		#endif
		__real_free(p);
	}
	// WARNING this breaks all assumptions C++ has, so it's probably best to call it in an inline assembly statement and call __crt0_exit in the next instruction
	// deallocate_exit does that
	static void __attribute__((used)) deallocate_all()
	{
		if (buckets != NULL)
		{
			for (uint32_t i = 0;i<buckets_size;i++)
			{
				if (buckets[i].list != NULL)
				{
					for (uint32_t a = 0;a<buckets[i].size;a++)
					{
						if (buckets[i].list[a] != NULL)
						{
							__real_free(buckets[i].list[a]);
						}
					}
					__real_free(buckets[i].list);
				}
			}
			__real_free(buckets);
		}
		#ifdef DEBUG
			printf("malloc count: %lu\n",malloc_count);
			fflush(stdout);
		#endif
	}
	void __attribute__ ((noreturn)) deallocate_exit(int code)
	{
		register int32_t c asm("r0") = code;
		asm volatile(
		"bl deallocate_all \n"
		"bl __crt0_exit"::"r" (c));
		__builtin_unreachable();
	}
	#pragma GCC reset_options
}